/*
 * Copyright (C) 2001 LAAS/CNRS 
 *

	     Written by T. Simeon 
	     july 91

  This package provides a set of functions for manipulating AVL trees.

  These trees are binary and equilibrated search trees for which 
  the insertion, deletion or search operations can be performed 
  in O(log n).
  
  The main routines are:

     (int)AVLInsertNode((void **)avl , (void *)key , fct_compar)
     (int)AVLDeleteNode((void **)avl , (void *)key , fct_compar)
  (void *)AVLFindNode  ((void *) avl , (void *)key , fct_compar)
  
  They are generalized from Knuth (6.2.3). All comparisons are done with
  a user-supplied routine. This routine is called with two arguments.
  The first one is a pointer to the new key, the other one is a pointer 
  to the current key. The routine must return -1, 0 or 1 according to
  whether the first argument is to be considered less than, equal to or
  greater than the second argument. The comparaison function need not
  compare every byte, so arbitrary data may be contained in the elements
  in addtion to the value being compared.

  In addition, the following routines are available:

  (int)AVLCheckBalance((void *)avl)
       returns TRUE if the tree is well balanced (only for debugging purpose)

  (int)AVLFindDeleteFirstNode((void *)avl)				      
       returns the best node of the tree

  (int)AVLHeight((void *)avl)
  (int)AVLWidth((void *)avl)
  (int)AVLLength((void *)avl)
      return the height,the number of leaves and the total number of
      nodes in the tree.

 *
 * $Log: avl.h,v $
 * Revision 1.1  2006/10/17 18:09:42  sjoyeux
 * Initial revision
 *
 * Revision 1.1  2001/07/11 18:05:40  mallet
 * re-ecriture de p3d
 *
 */
#ifndef AVL_H
#define AVL_H

/* ------------------------------------------------------------------- */

typedef struct AVLNode {
   struct AVLNode *left;
   struct AVLNode *right;
   void *key;
   short bal;
} AVLNode;

#define TRUE	1
#define FALSE	0

#define L_BAL -1
#define E_BAL 0
#define R_BAL 1

#define HEIGHT_NO_MODIFIED 0
#define HEIGHT_MODIFIED    1

/* ------------------------------------------------------------------- */

#define AVL_SET_LEFT(n,nl)  ((n)->left  = nl)
#define AVL_SET_RIGHT(n,nr) ((n)->right = (nr))
#define AVL_SET_KEY(n,k)    ((n)->key   = (k))
#define AVL_SET_BAL(n,b)    ((n)->bal   = (b))

#define AVL_GET_LEFT(n)     ((n)->left)
#define AVL_GET_RIGHT(n)    ((n)->right)
#define AVL_GET_BAL(n)      ((n)->bal)
#define AVL_GET_KEY(n)      ((n)->key)

#define AVL_ADD_LEFT(n,key)  (((n)->bal)--,(n)->left=makeNode(key))
#define AVL_ADD_RIGHT(n,key) (((n)->bal)++,(n)->right=makeNode(key))
#define AVL_ADD_BAL(n,val)   (((n)->bal) += val)
#define AVL_DEL_LEFT(n)      (((n)->bal)++,freeNode((n)->left),(n)->left=NULL)
#define AVL_DEL_RIGHT(n)     (((n)->bal)--,freeNode((n)->right),(n)->right=NULL)

#define AVL_BAL_EQUILIBRATED(bal) (abs(bal) > 1 ? FALSE : TRUE)

#define AVL_GET_SON(n,sens)     ((sens)==L_BAL ? (n)->left : (n)->right)
#define AVL_GET_ADRSON(n,sens)  ((sens)==L_BAL ? &(n)->left : &(n)->right)
#define AVL_GET_OTHERSON(n,sens)((sens)==L_BAL ? (n)->right : (n)->left)
#define AVL_ADD_SON(n,sens,key) ((sens)==L_BAL ? AVL_ADD_LEFT(n,key) :        \
                                                 AVL_ADD_RIGHT(n,key))
 

/* ------------------------------------------------------------------- */

int	AVLInsertNode(AVLNode **p_avl, void *key,
		int (*fct_order)(void *, void *));
int	AVLDeleteNode(AVLNode **p_avl, void *key,
		int (*fct_order)(void *, void *));

void *	AVLFindNode(AVLNode *avl, void *key,
		int (*fct_order)(void *, void *));
void *	AVLFindDeleteFirstNode(AVLNode **p_avl);
void *	AVLFindDeleteLastNode(AVLNode **p_avl);

int	AVLCheckBalance(AVLNode *avl);
int	AVLHeight(AVLNode *avl);
int	AVLWidth(AVLNode *avl);
int	AVLLength(AVLNode *avl);

#endif /* AVL_H */
