/**
 ** Thierry Peynot - Sat Mar 26 2005
 ** Copyright (c) 2005 LAAS/CNRS 
 **
 ** Functions to display useful information in GDHE
 **
 ** All functions names start with p3dDisp
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <t3d/t3d.h>
#include "libp3d.h"
/* #include "libp3dError.h" */
#include "p3dInt.h"

#include <gdhe/GDHE.h>


/*-- Global variables --*/
static float  dalatitoScale=0.5;
static int  displayConfigInit=0;  /* == 1 if the init for display config has been done */

/* Constants */

/*-- Correction du z Robot reel / Robot GDHE (ie ~= rayon des roues du modele GDHE de dala) ---- */
#define Z_DALA_GDHE 0.20


#define DEBUG 0


/**********************************************************************
 *
 * Create First waypoint to be displayed (as a red cylinder)
 *
 **********************************************************************/

int
p3dDispCreateFirstWayPoint(float  diameter,  /* diameter of the cylinder to display */
			   float  height)    /* height of that cylinder */
{
  char	 gdheCmdBuf[1024];

  sprintf (gdheCmdBuf, 
	   "set robots(waypt0) {object waypoint0 {color 255 0 0 ; cylinder 0 0 0.0 z %f %f} }", 
	   diameter, height );
  eval_expression (gdheCmdBuf);

  return 0;
}




/**********************************************************************
 *
 * Create a number N of waypoints to be displayed (as black cylinders)
 *
 **********************************************************************/

int
p3dDispCreateWayPoints(int    wayptNb,   /* Number of waypoints */
		       float  diameter,  /* diameter of the cylinders to display */
		       float  height)    /* height of those cylinders */
{
  char	 gdheCmdBuf[1024];
  int    i;

  for(i = 1; i <= wayptNb; i++)
    {
      /* waypoints numbers are from 1 to N */
      sprintf (gdheCmdBuf, 
	       "set robots(waypt%d) {object waypoint {color 0 0 0 ; cylinder 0 0 0.0 z %f %f} }", 
	       i, diameter, height );
      eval_expression (gdheCmdBuf);
    }

  return 0;
}



/**********************************************************************
 *
 * Send a new Position for waypoint number i
 *
 **********************************************************************/

int
p3dDispSetPosWaypoint(int     i,      /* waypoint number i */
		      float   x,      /* Waypoint position */
		      float   y )
{
  char	 gdheCmdBuf[1024];

  sprintf (gdheCmdBuf, "set pos(waypt%d) {0 0 0 %f %f 0}", i, x,y);
  eval_expression (gdheCmdBuf);

  return 0;
}





/**********************************************************************
 *
 * Send a new Position of the arc (illustrating Vref,Wref) to GDHE
 *
 **********************************************************************/

int
p3dDispSetPosArc(double  theta,  /* Robot Position (where the arc starts) */
	     double  x,
	     double  y,
	     double  z )
{
  char	 gdheCmdBuf[1024];

  /*-- Set z+0.2 to put the gdhe arc at the center of Robot --*/
  sprintf (gdheCmdBuf, 
	   "set pos(arc) {%lf 0 0 %lf %lf %lf}",
	   RAD_TO_DEG(theta), x, y, z+Z_DALA_GDHE);
  eval_expression (gdheCmdBuf);

  return 0;
}




/**********************************************************************
 *
 * Draw an arc to illustrate (Vref,Wref) -> "SpeedRefArc"
 *
 **********************************************************************/

int
p3dDispDrawSpeedRefArc(double  vref,
		       double  wref,
		       int     numNodes,  /* Number of nodes in the arc */
		       double  length)    /* length of the arc */
{
  char	 gdheCmdBuf[1024];
  char	 temp[1024];
  
  double   dtheta;
  double   Radius;
  double   xNode=0.0, yNode=0.0;  /* Arc nodes coordinate */
  
  int  i;
  
  if(wref != 0.0)
    {
      sprintf (gdheCmdBuf, 
	       "set robots(arc) {polyline %d ", numNodes);
      
      /* first Node of the arc */
      strcat(gdheCmdBuf, "0 0 0 ");

      Radius = vref / wref;
      dtheta = length / (Radius*numNodes);

      /* Calculates the nodes, angle should stay < Pi/2 = 1.57*/
      for(i = 1; i<numNodes; i++)
	{
	  if(i*dtheta < M_PI_2)
	    {
	      /* Arc Node coordinates in Robot frame */
	      xNode = Radius*sin((double) i * dtheta);
	      yNode = Radius*(1-cos((double) i * dtheta));
	    }
	  /* else, do nothing, don't put any more node */

	  sprintf (temp, "%lf %lf 0.0 ", xNode, yNode);
	  strcat(gdheCmdBuf, temp);
	}
      strcat(gdheCmdBuf, "}");
    }
  else
    sprintf (gdheCmdBuf, 
	     "set robots(arc) {polyline 2 0 0 0 %lf 0 0 }", length );
  
#if DEBUG
  /* Just to check */
  printf("p3dDisplay: %s\n", gdheCmdBuf);
  printf("p3dDisplay: %d nodes\n", numNodes);
#endif

  eval_expression (gdheCmdBuf);

  return 0;
}





/**********************************************************************
 *
 * Set Dala Position in GDHE
 *
 **********************************************************************/

int
p3dDispSetPosDala(double  theta,  /* Robot Position */
		  double  pitch,
		  double  roll,
		  double  x,
		  double  y,
		  double  z )
{
  char	 gdheCmdBuf[1024];

  /* NB: zRobotGDHE needs a correction?? as in GDHE zRobot=0 on the ground?? */
  sprintf (gdheCmdBuf, 
	   "set pos(dala) {%lf %lf %lf %lf %lf %lf}",
	   RAD_TO_DEG(theta), RAD_TO_DEG(pitch), RAD_TO_DEG(roll), x, y, z-Z_DALA_GDHE);
  eval_expression (gdheCmdBuf);


  return 0;
}





/**********************************************************************
 *
 * Create "Mini-Robots" to illustrate the predicted configurations
 *
 **********************************************************************/

int
p3dDispCreateMiniDalas(int    robotsNb,   /* Number of nodes in P3DArc (predicted config.) => nb of GDHE robots to create */
		       float  scale)  /* scale factor for MiniDalas's size */
/* normally, scale should be <= 1.0 */
{
  char	 gdheCmdBuf[1024];
  int    i;


  if(scale > 1.0)
    printf("p3dDispCreateMiniDalas: Warning: mini Dala scale is > 1.0\n");

  for(i = 0; i < robotsNb; i++)
    {
      sprintf (gdheCmdBuf, 
	       "set robots(dalatito%d) {object minidala {miniatrv %f} }",
	       i, scale);
      eval_expression (gdheCmdBuf);
    }

  dalatitoScale = scale;
  displayConfigInit = 1;

  return 0;
}




/* Convert a predicted configuration into a T3D (type=T3DEuler) */
/* returns that T3D */
T3D*
arcNodeConfigToT3D(P3DNode*    node)    /* node */
{
  T3D*  nodeToRobot;

  /* Read the position for the current arc node and put it in the T3D "nodeToRobot" */
  /*  (this is a relative position wrt the robot) */
  nodeToRobot = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  nodeToRobot->euler.euler[0] = node->theta;
  nodeToRobot->euler.euler[1] = 0.0;  /* pitch */
  nodeToRobot->euler.euler[2] = 0.0;
  nodeToRobot->euler.euler[3] = node->x;
  nodeToRobot->euler.euler[4] = node->y;
  nodeToRobot->euler.euler[5] = 0.0;

  return(nodeToRobot);
}









/**********************************************************************
 *
 * Displays a given predicted configuration of the robot(dala)
 *
 * Input: - the config position wrt the robot
 *        - pitch and roll predicted for that config.
 *        - the current position of the robot (mainToOrigin)
 *        - The No of the node which configuration is to be displayed
 *        - The scale of the mini robot to be displayed
 *
 **********************************************************************/

int
p3dDisplayOnePredictedConfig(T3D*  nodeToRobot,   /* Config pos relative to robot */
			     double     configPitch,      /* Pitch for that config */
			     double     configRoll,      /* Roll for that config */
			     double     configZ,       /* height */
			     T3D*  robotToOrigin,  /* Robot Current position */
			     int        nodeNb,
			     float      scale )  /* scale of mini-Dala */
{
  T3D    nodeToOrigin;
  char	 gdheCmdBuf[1024];
  
  int    k;
  
#if DEBUG
  printf("p3dCfDisplay: nodeToRobot = ");
  for(k=0; k<6; k++)
    printf("%f ", nodeToRobot->euler.euler[k]);
  printf("\n");

  printf("p3dCfDisplay: robotToOrigin = ");
  for(k=0; k<6; k++)
    printf("%f ", robotToOrigin->euler.euler[k]);
  printf("\n");
#endif

  /* Read the position for the current arc node */
  /*  (this is a relative position wrt the robot) */

  /* Compute node position to origin */
  t3dCompIn(&nodeToOrigin, nodeToRobot, robotToOrigin);
  t3dConvertTo(T3D_EULER, &nodeToOrigin);

  /* TODO: put the roll at the begining? (pas beau sinon...) */
#if DEBUG
  printf("p3dCfDisplay: nodeToOrigin = ");
  for(k=0; k<6; k++)
    printf("%f ", nodeToOrigin.euler.euler[k]);
  printf("\n");
#endif

  sprintf (gdheCmdBuf, 
	   "set pos(dalatito%d) {%lf %lf %lf %lf %lf %lf}", nodeNb,
	   RAD_TO_DEG(nodeToOrigin.euler.euler[0]),   /* theta */
	   RAD_TO_DEG(configPitch),   /* pitch */
	   RAD_TO_DEG(configRoll),   /* roll */
	   nodeToOrigin.euler.euler[3],               /* x */
	   nodeToOrigin.euler.euler[4],               /* y */
	   (configZ - Z_DALA_GDHE)*scale);    /** Z a verifier(scale?) **/
  eval_expression (gdheCmdBuf);

#if DEBUG
  printf("p3dDisplayConfig: %s\n", gdheCmdBuf);
#endif

  return 0;
}





/**********************************************************************
 *
 * Displays some predicted configurations of the robot(dala) on a P3DArc
 *
 * Input: - the P3DArc
 *        - Global position of robot
 *        - which nodes are to be displayed
 *
 **********************************************************************/

int 
p3dDisplayP3DArcConfig (P3DArc*  currentArc,
			T3D*  robotToOrigin,  /* Robot Position in global frame */
			int   firstNodeNb,
			int   lastNodeNb,
			int   interval )  /* interval between 2 succesive nodes displayed */
{
  T3D*   nodeToRobot;

  int   nodeNb;
#if DEBUG
  int   k;
#endif

  /* Check Creation of GDHE objects has been done */
  if( displayConfigInit == 0 )
    {
      fprintf(stderr, "p3dDisplay: Display Config Init has not been done\n");
      return -1;
    }

  /* Check params */
  if( (firstNodeNb >= currentArc->n) ||
      (lastNodeNb >= currentArc->n)  ||
      (firstNodeNb > lastNodeNb) ||
      interval <= 0 )
    {
      fprintf(stderr, "p3dDisplay: There are only %d in the arc\n", currentArc->n);
      return -1;
    }


#if DEBUG
  /* check position */
  printf("p3dDisplay: robotToOrigin = ");
  for(k=0; k<6; k++)
    printf("%f ", robotToOrigin->euler.euler[k]);
  printf("\n");
#endif

  for(nodeNb = firstNodeNb; 
      nodeNb <= lastNodeNb;
      nodeNb += interval )
    {
      nodeToRobot = arcNodeConfigToT3D(&(currentArc->nodes[nodeNb]));
      
#if DEBUG
      printf("p3dDisplay: nodeToRobot[%d] = ", nodeNb);
      for(k=0; k<6; k++)
	printf("%f ", nodeToRobot->euler.euler[k]);
      printf("\n");
#endif

      p3dDisplayOnePredictedConfig(nodeToRobot,
				   currentArc->nodes[nodeNb].conf.pitch,
				   currentArc->nodes[nodeNb].conf.roll,
				   currentArc->nodes[nodeNb].conf.z,
				   robotToOrigin,
				   nodeNb, dalatitoScale);
    }
  
  return 0;
}
